/*
 * Copyright (c) Ian F. Darwin 1986-1995.
 * Software written by Ian F. Darwin and others;
 * maintained 1995-present by Christos Zoulas and others.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, without modification,
 *    this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * fsmagic - magic based on filesystem info - directory, special files, etc.
 */

#include "file.h"
#include "magic.h"
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#include <sys/stat.h>
/* Since major is a function on SVR4, we cannot use `ifndef major'.  */
#ifdef MAJOR_IN_MKDEV
# include <sys/mkdev.h>
# define HAVE_MAJOR
#endif
#ifdef MAJOR_IN_SYSMACROS
# include <sys/sysmacros.h>
# define HAVE_MAJOR
#endif
#ifdef major			/* Might be defined in sys/types.h.  */
# define HAVE_MAJOR
#endif
  
#ifndef HAVE_MAJOR
# define major(dev)  (((dev) >> 8) & 0xff)
# define minor(dev)  ((dev) & 0xff)
#endif
#undef HAVE_MAJOR

#ifndef	lint
FILE_RCSID("@(#)$Id: fsmagic.c,v 1.46 2005/06/25 15:52:14 christos Exp $")
#endif	/* lint */

#include <fcntl.h>
#include <regex.h>

#define RE_DEVEL_LIBDIR "(/usr(/X11R6)?)?/lib(|64)(/gcc(-lib)?/.+)?/"
static const char re_devel_symlink[] = RE_DEVEL_LIBDIR "[^/]+\\.so$";
static const char re_devel_archive[] = RE_DEVEL_LIBDIR "[^/]+\\.a$";

protected int
is_devel_symlink(const char *fn)
{
    enum {
	RE_STATUS_INIT,
	RE_STATUS_ERROR,
	RE_STATUS_RUN
    };
    static int re_status = RE_STATUS_INIT;
    static regex_t re;
    int err;

    switch (re_status) {
    case RE_STATUS_INIT:
	err = regcomp(&re, re_devel_symlink, REG_EXTENDED);
	if (err != REG_NOERROR) {
	    re_status = RE_STATUS_ERROR;
	    return 0;
	}
	/* fall through */
    case RE_STATUS_RUN:
    {
	regmatch_t pmatch[1];
	err = regexec(&re, fn, sizeof(pmatch)/sizeof(pmatch[0]), pmatch, 0);
	return err == 0;
    }
    case RE_STATUS_ERROR:
    default:
	return 0;
    }
}

protected int
is_devel_archive(const char *fn)
{
    enum {
	RE_STATUS_INIT,
	RE_STATUS_ERROR,
	RE_STATUS_RUN
    };
    static int re_status = RE_STATUS_INIT;
    static regex_t re;
    int err;

    switch (re_status) {
    case RE_STATUS_INIT:
	err = regcomp(&re, re_devel_archive, REG_EXTENDED);
	if (err != REG_NOERROR) {
	    re_status = RE_STATUS_ERROR;
	    return 0;
	}
	/* fall through */
    case RE_STATUS_RUN:
    {
	regmatch_t pmatch[1];
	err = regexec(&re, fn, sizeof(pmatch)/sizeof(pmatch[0]), pmatch, 0);
	return err == 0;
    }
    case RE_STATUS_ERROR:
    default:
	return 0;
    }
}

/* Static archive parser derived from X.org code.  */
#define ARMAG	"!<arch>\n"
#define SARMAG	8

struct ar_hdr {
    char ar_name[16];
    char ar_date[12];
    char ar_uid[6];
    char ar_gid[6];
    char ar_mode[8];
    char ar_size[10];
    char ar_fmag[2];
};

#define ELFMAG	"\177ELF"
#define SELFMAG	4

#define EI_NIDENT	16
#define EI_CLASS	4
#define ELFCLASS32	1
#define ELFCLASS64	2

protected int
classify_archive_fd(int arfd)
{
    unsigned char armagic[SARMAG];
    unsigned char e_ident[EI_NIDENT];
    struct ar_hdr hdr;
    unsigned int size;
    unsigned int offset;

    /* Check archive magic.  */
    read(arfd, armagic, SARMAG);
    if (strncmp(armagic, ARMAG, SARMAG) != 0)
	return -1;

    /* Sequentially read all the contents until an object is found.  */
    while (read(arfd, &hdr, sizeof(struct ar_hdr))) {
	sscanf(hdr.ar_size, "%u", &size);
	offset = lseek(arfd, 0, SEEK_CUR);

	/* Skip Symbol Table.  */
	if ((hdr.ar_name[0] == '/' && hdr.ar_name[1] == ' ') ||
	    strncmp(hdr.ar_name, "__.SYMDEF", 9) == 0) {
	    /* If the file name is NULL, then it is a symbol table */
	    offset = lseek(arfd, offset + size, SEEK_SET);
	    if (offset & 0x1)	/* odd value */
		offset = lseek(arfd, 1, SEEK_CUR);	/* make it an even boundary */
	    continue;
	}

	/* Skip String Table.  */
	if (hdr.ar_name[0] == '/' && hdr.ar_name[1] == '/') {
	    /* If the file name is '/', then it is a string table */
	    offset = lseek(arfd, size, SEEK_CUR);
	    if (offset & 0x1)	/* odd value */
		offset = lseek(arfd, 1, SEEK_CUR);	/* make it an even boundary */
	    continue;
	}

	/* Reajust offset for BSD 4.4 style long member name.  */
	if (hdr.ar_name[0] == '#' && hdr.ar_name[1] == '1' &&
	    hdr.ar_name[2] == '/') {
	    int namesz;
	    if (sscanf(&hdr.ar_name[3], "%d", &namesz) != 1)
		return -1;
	    offset += namesz;
	    size -= namesz;
	}

	/* Check for an ELF object.  */
	read(arfd, e_ident, sizeof(e_ident));
	if (strncmp(&e_ident[0], ELFMAG, SELFMAG) == 0) {
	    switch (e_ident[EI_CLASS]) {
	    case ELFCLASS32: return 32;
	    case ELFCLASS64: return 64;
	    }
	}

	offset = lseek(arfd, offset + size, SEEK_SET);
	if (offset & 0x1)	/* odd value */
	    lseek(arfd, 1, SEEK_CUR);	/* make it an even boundary */
  }

  return -1;
}

protected int
classify_archive(const char *archive_filename)
{
    int elf_class;
    int fd;

    if ((fd = open(archive_filename, O_RDONLY)) < 0)
	return -1;

    elf_class = classify_archive_fd(fd);
    close(fd);
    return elf_class;
}

protected int
file_fsmagic(struct magic_set *ms, const char *fn, struct stat *sb, int fdepth)
{
	int ret = 0;
#ifdef	S_IFLNK
	char buf[BUFSIZ+4];
	char buf2[BUFSIZ+BUFSIZ+4];
	char *dbuf = buf;
	int nch;
	struct stat tstatbuf;
#endif

	if (fn == NULL)
		return 0;

	/*
	 * Fstat is cheaper but fails for files you don't have read perms on.
	 * On 4.2BSD and similar systems, use lstat() to identify symlinks.
	 */
#ifdef	S_IFLNK
	if ((ms->flags & MAGIC_SYMLINK) == 0)
		ret = lstat(fn, sb);
	else
#endif
	ret = stat(fn, sb);	/* don't merge into if; see "ret =" above */

	if (ret) {
		if (ms->flags & MAGIC_ERROR) {
			file_error(ms, errno, "cannot stat `%s'", fn);
			return -1;
		}
		if (file_printf(ms, "cannot open `%s' (%s)",
		    fn, strerror(errno)) == -1)
			return -1;
		ms->haderr++;
		return -1;
	}

	if ((ms->flags & MAGIC_MIME) != 0) {
		if ((sb->st_mode & S_IFMT) != S_IFREG) {
			if (file_printf(ms, "application/x-not-regular-file")
			    == -1)
				return -1;
			return 1;
		}
	}
	else {
#ifdef S_ISUID
		if (sb->st_mode & S_ISUID) 
			if (file_printf(ms, "setuid ") == -1)
				return -1;
#endif
#ifdef S_ISGID
		if (sb->st_mode & S_ISGID) 
			if (file_printf(ms, "setgid ") == -1)
				return -1;
#endif
#ifdef S_ISVTX
		if (sb->st_mode & S_ISVTX) 
			if (file_printf(ms, "sticky ") == -1)
				return -1;
#endif
	}
	
	switch (sb->st_mode & S_IFMT) {
	case S_IFDIR:
		if (file_printf(ms, "directory") == -1)
			return -1;
		return 1;
#ifdef S_IFCHR
	case S_IFCHR:
		/* 
		 * If -s has been specified, treat character special files
		 * like ordinary files.  Otherwise, just report that they
		 * are block special files and go on to the next file.
		 */
		if ((ms->flags & MAGIC_DEVICES) != 0)
			break;
#ifdef HAVE_ST_RDEV
# ifdef dv_unit
		if (file_printf(ms, "character special (%d/%d/%d)",
		    major(sb->st_rdev), dv_unit(sb->st_rdev),
		    dv_subunit(sb->st_rdev)) == -1)
			return -1;
# else
		if (file_printf(ms, "character special (%ld/%ld)",
		    (long) major(sb->st_rdev), (long) minor(sb->st_rdev)) == -1)
			return -1;
# endif
#else
		if (file_printf(ms, "character special") == -1)
			return -1;
#endif
		return 1;
#endif
#ifdef S_IFBLK
	case S_IFBLK:
		/* 
		 * If -s has been specified, treat block special files
		 * like ordinary files.  Otherwise, just report that they
		 * are block special files and go on to the next file.
		 */
		if ((ms->flags & MAGIC_DEVICES) != 0)
			break;
#ifdef HAVE_ST_RDEV
# ifdef dv_unit
		if (file_printf(ms, "block special (%d/%d/%d)",
		    major(sb->st_rdev), dv_unit(sb->st_rdev),
		    dv_subunit(sb->st_rdev)) == -1)
			return -1;
# else
		if (file_printf(ms, "block special (%ld/%ld)",
		    (long)major(sb->st_rdev), (long)minor(sb->st_rdev)) == -1)
			return -1;
# endif
#else
		if (file_printf(ms, "block special") == -1)
			return -1;
#endif
		return 1;
#endif
	/* TODO add code to handle V7 MUX and Blit MUX files */
#ifdef	S_IFIFO
	case S_IFIFO:
		if((ms->flags & MAGIC_DEVICES) != 0)
			break;
		if (file_printf(ms, "fifo (named pipe)") == -1)
			return -1;
		return 1;
#endif
#ifdef	S_IFDOOR
	case S_IFDOOR:
		if (file_printf(ms, "door") == -1)
			return -1;
		return 1;
#endif
#ifdef	S_IFLNK
	case S_IFLNK:
		if ((nch = readlink(fn, buf, BUFSIZ-1)) <= 0) {
			if (ms->flags & MAGIC_ERROR) {
			    file_error(ms, errno, "unreadable symlink `%s'",
				fn);
			    return -1;
			}
			if (file_printf(ms,
			    "unreadable symlink `%s' (%s)", fn,
			    strerror(errno)) == -1)
				return -1;
			return 1;
		}
		buf[nch] = '\0';	/* readlink(2) forgets this */

		/* If broken symlink, say so and quit early. */
		if (*buf == '/') {
		    if (stat(buf, &tstatbuf) < 0) {
			    if (ms->flags & MAGIC_ERROR) {
				    file_error(ms, errno, 
					"broken symbolic link to `%s'", buf);
				    return -1;
			    } 
			    if (file_printf(ms, "broken symbolic link to `%s'",
				buf) == -1)
				    return -1;
			    return 1;
		    }
		}
		else {
			char *tmp;

			if ((tmp = strrchr(fn,  '/')) == NULL) {
				tmp = buf; /* in current directory anyway */
			} else {
				if (tmp - fn + 1 > BUFSIZ) {
					if (ms->flags & MAGIC_ERROR) {
						file_error(ms, 0, 
						    "path too long: `%s'", buf);
						return -1;
					}
					if (file_printf(ms,
					    "path too long: `%s'", fn) == -1)
						return -1;
					return 1;
				}
				(void)strcpy(buf2, fn);  /* take dir part */
				buf2[tmp - fn + 1] = '\0';
				(void)strcat(buf2, buf); /* plus (rel) link */
				tmp = buf2;
			}
			if (stat(tmp, &tstatbuf) < 0) {
				if (ms->flags & MAGIC_ERROR) {
					file_error(ms, errno, 
					    "broken symbolic link to `%s'",
					    buf);
					return -1;
				}
				if (file_printf(ms,
				    "broken symbolic link to `%s'", buf) == -1)
					return -1;
				return 1;
			}
			dbuf = tmp;
		}

		/* Otherwise, handle it. */
		if ((ms->flags & MAGIC_SYMLINK) != 0) {
			const char *p;
			ms->flags &= MAGIC_SYMLINK;
			p = magic_do_file(ms, buf, fdepth + 1);
			ms->flags |= MAGIC_SYMLINK;
			return p != NULL ? 1 : -1;
		} else if ((ms->flags & MAGIC_DEVEL_SYMLINK) != 0) {
			const char *p;
			ms->flags &= MAGIC_DEVEL_SYMLINK;
			p = magic_do_file(ms, dbuf, fdepth + 1);
			ms->flags |= MAGIC_DEVEL_SYMLINK;
			return p != NULL ? 1 : -1;
		} else { /* just print what it points to */
			/* special output for -devel .so symlinks */
			if (fdepth == 0 && is_devel_symlink(fn)) {
				const char *p;
				if (file_printf(ms, "symbolic link to ") == -1)
					return -1;
				ms->flags |= MAGIC_DEVEL_SYMLINK;
				p = magic_do_file(ms, dbuf, fdepth + 1);
				ms->flags &= ~MAGIC_DEVEL_SYMLINK;
				return p != NULL ? 1 : -1;
			}
			if (file_printf(ms, "symbolic link to `%s'",
			    buf) == -1)
				return -1;
		}
	return 1;
#endif
#ifdef	S_IFSOCK
#ifndef __COHERENT__
	case S_IFSOCK:
		if (file_printf(ms, "socket") == -1)
			return -1;
		return 1;
#endif
#endif
	case S_IFREG:
		break;
	default:
		file_error(ms, 0, "invalid mode 0%o", sb->st_mode);
		return -1;
		/*@notreached@*/ break;
	}

	/*
	 * regular file, check next possibility
	 *
	 * If stat() tells us the file has zero length, report here that
	 * the file is empty, so we can skip all the work of opening and 
	 * reading the file.
	 * But if the -s option has been given, we skip this optimization,
	 * since on some systems, stat() reports zero size for raw disk
	 * partitions.  (If the block special device really has zero length,
	 * the fact that it is empty will be detected and reported correctly
	 * when we read the file.)
	 */
	if ((ms->flags & MAGIC_DEVICES) == 0 && sb->st_size == 0) {
		if (file_printf(ms, (ms->flags & MAGIC_MIME) ?
		    "application/x-empty" : "empty") == -1)
			return -1;
		return 1;
	}

	/* Check for static archives in usual development libdirs.  */
	if (is_devel_archive(fn)) {
		int ar_class = classify_archive(fn);
		if (ar_class == 32 || ar_class == 64) {
			if (file_printf(ms, "ar archive with %d-bit ELF objects", ar_class) == -1)
				return -1;
			return 1;
		}
	}
	return 0;
}
